/*
   alt.c -- PIKT alarm script processing routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/alt.c,v 1.61 2007/03/12 09:59:07 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2007 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktd.h"

char altline[MAXBUF+1];

int freevarslot(void);

/* register an alarm: initialize its data structures; add it to the alarms
   list
 */
void
regalm (char *name)
{
	static char fnnam[] = "regalm";
	alrm *nalp;
	register int i, j;

	nalp = (alrm *) mymalloc(sizeof(alrm), fnnam);
	nalp->next = (alrm *) NULL;
	nalp->name = name;
	nalp->status = ACTIVE;	/* if we set this to NONE here, where later
				   do we set this to ACTIVE for command-line
				   ('pikt +C ...') operations?
				 */
	nalp->level = NIL;
	nalp->task = mystrdup(nilstr, fnnam);
	nalp->input.type = 0;
	nalp->input.source = (char *) NULL;
	nalp->input.tmpfil = (char *) NULL;
	nalp->input.bakfil = (char *) NULL;
	nalp->input.logname = (char *) NULL;
	nalp->input.logino = 0L;
	nalp->input.logsize = 0L;
	nalp->input.logposend = 0L;
	nalp->filter = (char *) NULL;
	nalp->separators = (char *) NULL;
	nalp->data.datcnt = 0;
	nalp->data.maxdatndx = 0;
	nalp->data.postype = 0;
	nalp->data.regexp = (char *) NULL;
	nalp->keys = (ky *) NULL;
	nalp->begin = (rul *) NULL;
	nalp->rules = (rul *) NULL;
	nalp->end = (rul *) NULL;
	for (i=0; i<=MAXVARS; i++) {
		nalp->vars[i] = (var *) mymalloc(sizeof(var), fnnam);
		(nalp->vars[i])->name = (char *) NULL;
		(nalp->vars[i])->valtype = NIL;
		(nalp->vars[i])->dattype = NIL;
		(nalp->vars[i])->arytype = NIL;
		(nalp->vars[i])->arydim = NIL;
		(nalp->vars[i])->oldhst  = FALSE;
		(nalp->vars[i])->newhst  = FALSE;
		for (j=0; j<=9; j++) {
			(nalp->vars[i])->datspc[j] = (spc *) NULL;
			(nalp->vars[i])->nodes[j] = 0;
		}
		(nalp->vars[i])->direction = FORWARD;
		(nalp->vars[i])->arkeys = (ky *) NULL;
		(nalp->vars[i])->sarkeys = (ky *) NULL;
		(nalp->vars[i])->rarkeys = (ky *) NULL;
		(nalp->vars[i])->listindex = NIL;
		(nalp->vars[i])->keysfn = FALSE;
		(nalp->vars[i])->sortkeysfn = FALSE;
		(nalp->vars[i])->mapkeyscnt = 0;
		(nalp->vars[i])->mapkeysvrndxary = NIL;
	}
	nalp->vrndxhi = 0;
	nalp->scvars = (ky *) NULL;
	nalp->valhst = (val *) NULL;
	initvals(&(nalp->valhst));
	nalp->hsttyp = 0;
	nalp->hsterr = TRUE;		/* guilty until proven innocent */
	nalp->history = (hist *) NULL;
	nalp->rewriteinfil = FALSE;
	if (! alarms)
		alarms = nalp;
	else
		alarmcurr->next = nalp;
	alarmcurr = nalp;
	return;
}

/* register a variable: if in a dat statement, process its positional info;
   initialize its data structures; set its valtype, dattype, arytype, arydim,
   arndx & arkey; return the var's vrndx
 */
int
regvar (char *name, int valtype, int dattype, int arytype, int arydim, signed long *arndx, char *arkey, char *position)
{
	static char fnnam[] = "regvar";
	char *cp;
	int i, n;
	signed int vrndx;
	#define MAXPOS 16
	char pos[MAXPOS+1];
	int pl;
	spc *sp;
	int trimtyp;

	trimtyp = NEITHER;
	vrndx = varndx(name);
	if (position) {		/* hence, a var in a dat statement */
		if ((pl = strlen(position)) > MAXPOS) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, data position too long:  %s", ALMNAM, position);
			STATUS = ERROR;
			return ERROR;
		}
		/* check if listed in previous dat statement */
		switch (vrndx) {
			case NIL:
				break;
			case INLINNOW:				
				/* we permit use of $inlin, $inline,
				   $inputline in dat statements
			 	 */
				/* can't think of a good way to check this;
				   can't do this:
				sp = lookupspc(vrndx, arndx, arkey);
				if (sp) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_ALREADY_SPECIFIED, __FILE__, __LINE__, fnnam, FALSE, "in %s, var %c%s%s already specified in previous dat statement", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
					STATUS = ERROR;
					return ERROR;
				}
				  because vrndx < 0 (INLINNOW < 0), which
				  would crash lookupspc()
				 */
				break;
			/* varndx() never returns INLINPRC
			case INLINPRC:
			 */
			case OUTLINNOW:
			/* varndx() never returns OUTLINPRC
			case OUTLINPRC:
			 */
			case RDLINNOW:
			/* varndx() never returns RDLINPRC
			case RDLINPRC:
			 */
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s is a reserved var name", ALMNAM, name);
				STATUS = ERROR;
				return ERROR;
				break;
			case VLPRS:
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s is a reserved var name", ALMNAM, name);
				STATUS = ERROR;
				return ERROR;
				break;
			case ERROR:
				logreturn(PIKTLOG, LOG_ERR, ERRID_FN_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempted use of reserved fn name %s as var", ALMNAM, name);
				STATUS = ERROR;
				return ERROR;
				break;
			default:
				/* find the datspc for this vrndx arndx[]/arkey, if any */
				sp = lookupspc(vrndx, arndx, arkey);
				if (sp) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_ALREADY_SPECIFIED, __FILE__, __LINE__, fnnam, FALSE, "in %s, var %c%s%s already specified in previous dat statement", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
					STATUS = ERROR;
					return ERROR;
				}
				break;
		}
		if (*position == '[') {
			if (*(position+pl-1) == ']') {
				*(position+pl-1) = '\0';
				strncpy(pos, position+1, MAXPOS);
				trimtyp = BOTH;
			} else
				POSTYP = ERROR;
		} else if (strchr(position,']')) {
				POSTYP = ERROR;
		} else {
			strncpy(pos, position, MAXPOS);
			trimtyp = NEITHER;
		}
		if (POSTYP == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, improperly formed data position:  %s", ALMNAM, position);
			STATUS = ERROR;
			return ERROR;
		}
		if ((cp = strchr(pos,','))) {
			if (! POSTYP)
				POSTYP = COLUMNS;
			else if (POSTYP != COLUMNS) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent data position: %s", ALMNAM, pos);
				STATUS = ERROR;
				return ERROR;
			}
		} else if (isdigit((int)pos[0]) || (pos[0] == '$')) {
			if (! POSTYP)
				POSTYP = FIELDS;
			else if (POSTYP != FIELDS) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent data position: %s", ALMNAM, pos);
				STATUS = ERROR;
				return ERROR;
			}
		} else {
			logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
			STATUS = ERROR;
			return ERROR;
		}
		/* at this point, have unique var name, valid pos */
		DATCNT++;
		if ((vrndx == NIL) || (vrndx == INLINNOW))
			vrndx = ++VRNDXHI;
		/* check if max var count exceeded */
		if (VRNDXHI > MAXVARS) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, maximum var count (%d) exceeded for var %c%s%s", ALMNAM, MAXVARS, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
			STATUS = ERROR;
			return ERROR;
		}
		/* check to see if data slot already taken */
		if (VARNAM(vrndx) && (arndx[1] == 0) && (! arkey)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_SLOT_ALREADY_TAKEN, __FILE__, __LINE__, fnnam, FALSE, "in %s, data slot %d already taken", ALMNAM, vrndx);
			STATUS = ERROR;
			return ERROR;
		}
		if (ARYTYP(vrndx) == NIL)
			ARYTYP(vrndx) = arytype;
		if (ARYDIM(vrndx) == NIL)
			ARYDIM(vrndx) = arydim;
		if (! (sp = regspc(vrndx, arndx, arkey))) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_INDEX_NULL, __FILE__, __LINE__, fnnam, FALSE, "in %s, for var %c%s[], null array index", ALMNAM, dattype ? '#' : '$', name);
			STATUS = ERROR;
			return ERROR;
		}
		sp->trim = trimtyp;
		if (POSTYP == COLUMNS) {
			*cp = '\0';
			if ((i = atoi(pos)) && (i > 0) && (i <= MAXLINE)) {
				sp->begincol = i;
				if (! strcmp(cp+1,"$")) {
					sp->endcol = MAXLINE;
				} else if ((i = atoi(cp+1)) && (i > 0) && (i <= MAXLINE)) {
					if (i >= sp->begincol)
						sp->endcol = i;
					else {
						logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
						STATUS = ERROR;
						return ERROR;
					}
				} else {
					logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
					STATUS = ERROR;
					return ERROR;
				}
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
				STATUS = ERROR;
				return ERROR;
			}
			regscvar(vrndx, arndx, arkey);
		} else if (pos[0] == '$') {
			if (pos[1]) {
				if (pos[1] == '-') {
					errno = 0;
					n = atoi(pos+2);
					if (errno) {
						logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
						STATUS = ERROR;
						return ERROR;
					}
					sp->fldnum = -(n+1);
				} else {
					logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: $s", ALMNAM, pos);
					STATUS = ERROR;
					return ERROR;
				}
			} else
				sp->fldnum = -1;
			MAXDATNDX = maxval(MAXDATNDX, -(sp->fldnum));
		} else if (isdigit((int)pos[0])) {
			errno = 0;
			n = atoi(pos);
			if (errno) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_DATA_POSITION_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad data position: %s", ALMNAM, pos);
				STATUS = ERROR;
				return ERROR;
			}
			sp->fldnum = n;
			MAXDATNDX = maxval(MAXDATNDX, sp->fldnum);
		}
	/* else in the remaining cases, no position, hence a var not in
	   a dat statement, e.g., in a set or for statement, or a
	   file handle
	 */
	} else {
		vrndx = varndx(name);
		if (stmcurr)
			stmcurr->vrndx = vrndx;
		if ((vrndx == NIL) || ((vrndx > 0) && (DATTYP(vrndx) == NIL))) {
			if (vrndx == NIL) {
				VRNDXHI = vrndx = freevarslot();
				if (stmcurr)
					stmcurr->vrndx = vrndx;
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, maximum var count (%d) exceeded for var %c%s%s", ALMNAM, MAXVARS, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
					STATUS = ERROR;
					return ERROR;
				}
			}
		} else if (vrndx == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FN_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempted use of reserved fn name %s as var", ALMNAM, name);
			STATUS = ERROR;
			return ERROR;
		} else if ((vrndx == RDLINNOW) || (vrndx == INLINNOW)) {
			if (doregvar) {	/* e.g., #split($inline,...) */
				return vrndx;
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, %s is a reserved var name", ALMNAM, name);
				STATUS = ERROR;
				return ERROR;
			}
		} else if (vrndx == OUTLINNOW) {
			return vrndx;
		} else { /* valid var name seen before */
			VALTYP(vrndx) |= valtype;
			FREE(arkey);
			return vrndx;
		}
	}
	if (vrndx <= 0)	/* NIL, INLINNOW, OUTLINNOW, RDLINNOW */
		vrndx = VRNDXHI;
	if (VARNAM(vrndx)) {
		if (strcmp(VARNAM(vrndx), name)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAMES_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent var names for vrndx %d: %s %s", ALMNAM, vrndx, VARNAM(vrndx), name);
			STATUS = ERROR;
			return ERROR;
		}
	} else
		VARNAM(vrndx) = mystrdup(name, fnnam);
	if (VALTYP(vrndx) == NIL)
		VALTYP(vrndx) = valtype;
	else if ((VALTYP(vrndx) & FIXEDVAL) && (VALTYP(vrndx) != valtype)) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_VAL_TYPES_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent val type for %c%s%s", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
		STATUS = ERROR;
		return ERROR;
	} else	/* should never be reached */
		VALTYP(vrndx) |= valtype;
	if (ARYDIM(vrndx) == NIL) {
		ARYDIM(vrndx) = arydim;
	} else {
		/* saw, e.g., $f before, $f[] now, or vice-versa */
		if (ARYDIM(vrndx) != arydim) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_SCALAR_ARRAY_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, scalar/array inconsistency for %c%s%s", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
			STATUS = ERROR;
			return ERROR;
		}
	}
	if (DATTYP(vrndx) == NIL)
		DATTYP(vrndx) = dattype;
	/* saw, e.g., $f[] before, #f[] now, or vice-versa */
	else if (DATTYP(vrndx) != dattype) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_DAT_TYPE_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent dat type for %c%s%s", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
		STATUS = ERROR;
		return ERROR;
	}
	if (ARYTYP(vrndx) == NIL)
		ARYTYP(vrndx) = arytype;
	/* saw, e.g., $f[2] before, $f["foo"] now, or vice-versa */
	else if (ARYTYP(vrndx) != arytype) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_ARRAY_TYPE_INCONSISTENT, __FILE__, __LINE__, fnnam, FALSE, "in %s, inconsistent array type for %c%s%s", ALMNAM, dattype ? '#' : '$', name, (arndx[1] || arkey) ? "[]" : nullstring);
		STATUS = ERROR;
		return ERROR;
	}
	return vrndx;
}

/* register a scancolumn var
 */
void
regscvar (int vrndx, signed long *arndx, char *arkey)
{
	static char fnnam[] = "regscvar";
	ky *nkp;
	register int k;

	nkp = (ky *) mymalloc(sizeof(ky), fnnam);
	nkp->prev = (ky *) NULL;
	nkp->next = (ky *) NULL;
	nkp->vrndx = vrndx;
	for (k=0; k<=MAXARYDIM; k++)
		nkp->arndx[k] = arndx[k];
	nkp->arkey = (((ARYTYP(vrndx) == STRING) && arkey) ? (char *) mystrdup(arkey, fnnam) : (char *)NULL);
	if (! SCVARS)
		SCVARS = nkp;
	else
		scvarcurr->next = nkp;
	scvarcurr = nkp;
	return;
}

/* register a history lookup key
 */
void
regkey (char *key, int valtype, signed long *arndx, char *arkey)
{
	static char fnnam[] = "regkey";
	ky *nkp;
	register int k;

	nkp = (ky *) mymalloc(sizeof(ky), fnnam);
	nkp->prev = (ky *) NULL;
	nkp->next = (ky *) NULL;
	switch ((nkp->vrndx = varndx(key+1))) {
		case NIL:
			logreturn(PIKTLOG, LOG_ERR, ERRID_KEY_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown key %s%s", ALMNAM, key, (arndx[1] || arkey) ? "[]" : nullstring);
			STATUS = ERROR;
			return;
			break;
		case INLINNOW:
			break;
		/* varndx() never returns INLINPRC
		case INLINPRC:
			break;
		 */
		case OUTLINNOW:
		/* varndx() never returns OUTLINPRC
		case OUTLINPRC:
		 */
		case RDLINNOW:
		/* varndx() never returns RDLINPRC
		case RDLINPRC:
		 */
			logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempted use of reserved var name %s as key", ALMNAM, key);
			STATUS = ERROR;
			return;
			break;
		case ERROR:
			logreturn(PIKTLOG, LOG_ERR, ERRID_FN_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempted use of reserved fn name %s as key", ALMNAM, key);
			STATUS = ERROR;
			return;
			break;
		default:
			if (! lookupspc(nkp->vrndx, arndx, arkey)) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_KEY_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown key %s%s", ALMNAM, key, (arndx[1] || arkey) ? "[]" : nullstring);
				STATUS = ERROR;
				return;
			}
			break;
	}
	dohistory = TRUE;		/* so, process history file */
	HSTTYP |= valtype;
	for (k=0; k<=MAXARYDIM; k++)
		nkp->arndx[k] = arndx[k];
	nkp->arkey = (arkey ? (char *) mystrdup(arkey, fnnam) : arkey);
	NEWHST(nkp->vrndx) = TRUE;
	if (! KEYS)
		KEYS = nkp;
	else
		keycurr->next = nkp;
	keycurr = nkp;	
	return;
}

/* register a rule
 */
void
regrul (int type)
{
	static char fnnam[] = "regrul";
	rul *nrp;

	nrp = (rul *) mymalloc(sizeof(rul), fnnam);
	nrp->next = (rul *) NULL;
	nrp->statements = (stm *) NULL;
	nrp->rulnum = NIL;
	switch (type) {
		case BEG:
			nrp->rulnum = 0;
			alarmcurr->begin = nrp;
			break;
		case RUL:
			if (! alarmcurr->rules) {
				nrp->rulnum = 1;
				alarmcurr->rules = nrp;
			} else {
				nrp->rulnum = rulcurr->rulnum + 1;
				rulcurr->next = nrp;
			}
			break;
		case END:
			if (! alarmcurr->rules) {
				nrp->rulnum = 1;
			} else {
				nrp->rulnum = rulcurr->rulnum + 1;
			}
			alarmcurr->end = nrp;
			break;
	}
	rulcurr = nrp;
	return;
}

/* register a token; if a var, replace its type char ($#?%@) with a signed
   char value indicating its type; replace the char immediately following
   the type char with an index from 1 to 255; so, for example, $foo might
   become the char sequence

      -1            4            o            o
       ^            ^            ^            ^
     VLNOW        4th var    unchanged    unchanged

   add the token to either the lhsexpr or rshexpr token list for the
   current statement
 */
void
regtok (signed char *token, int isfn)
{
	static char fnnam[] = "regtok";
	tok *ntp;
	signed int vrndx;
	int ndx;

	vrndx = NIL;
	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, ERRID_DEBUG, __FILE__, __LINE__, fnnam, FALSE, "regtok'ing %s", token);
	if (! isfn) {
		if ((*token == '$') || (*token == '#') || (*token == '?')) {
			if (isalnum((int)*(token+1))) {
				vrndx = varndx(token+1);
				switch (vrndx) {
					case VLSTR:
						break;
					case VLPRS:
						ndx = atoi(token+1);
						if (ndx < 0) {
							logreturn(PIKTLOG, LOG_ERR, ERRID_VALUE_OUT_OF_BOUNDS, __FILE__, __LINE__, fnnam, FALSE, "in %s, out-of-bounds value: %s", ALMNAM, token);
							STATUS = ERROR;
							return;
						} else {
							*token = VLPRS;
							*(token+1) = (unsigned char) ndx;
						}
						break;
					case INLINNOW:
						*token = INLINNOW;
						break;
					case OUTLINNOW:
						*token = OUTLINNOW;
						break;
					case RDLINNOW:
						*token = RDLINNOW;
						break;
					case ERROR:
					case NIL:
						/* error reported below */
						break;
					default:
						*token = VLNOW;
						*(token+1) = (unsigned char) vrndx;
						break;
				}
			}
		} else if ((*token == '%') && isalpha((int)*(token+1))) {
			vrndx = varndx(token+1);
			if (vrndx == NIL) {
				vrndx = freevarslot();
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, maximum var count (%d) exceeded for var %%%s%s", ALMNAM, MAXVARS, token, *(token+2) == '\[' ? "[]" : nullstring);
					STATUS = ERROR;
					return;
				} else
					VARNAM(vrndx) = mystrdup(token+1, fnnam);
			}
			switch (vrndx) {
				case VLSTR:
					break;
				case VLPRS:
				case INLINNOW:
				case OUTLINNOW:
				case RDLINNOW:
					logreturn(PIKTLOG, LOG_ERR, ERRID_HISTORY_REFERENCE_IMPROPER, __FILE__, __LINE__, fnnam, FALSE, "in %s, improper history reference for var %%%s", ALMNAM, token);
					STATUS = ERROR;
					return;
					break;
				case ERROR:
				case NIL:
					/* error reported below */
					break;
				default:
					*token = VLHST;
					*(token+1) = (unsigned char) vrndx;
					reghst(vrndx);
					break;
			}
		} else if ((*token == '@') && isalpha((int)*(token+1))) {
			vrndx = varndx(token+1);
			if (vrndx == NIL) {
				vrndx = freevarslot();
				if (vrndx == NIL) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_COUNT_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, maximum var count (%d) exceeded for var @%s%s", ALMNAM, MAXVARS, token, *(token+2) == '\[' ? "[]" : nullstring);
					STATUS = ERROR;
					return;
				} else
					VARNAM(vrndx) = mystrdup(token+1, fnnam);
			}
			switch (vrndx) {
				case VLSTR:
					break;
				case VLPRS:
				case ERROR:
				case NIL:
					/* error reported below */
					break;
				case INLINNOW:
					*token = INLINPRC;
					break;
				case OUTLINNOW:
					*token = OUTLINPRC;
					break;
				case RDLINNOW:
					*token = RDLINPRC;
					break;
				default:
					*token = VLPRC;
					*(token+1) = (unsigned char) vrndx;
					break;
			}
		} else if ((*token == '\[') && tokcurr && ! strcmp(tokcurr->token,"$")) {
			/* ah, so this is an instance of $[...]; set the
			   first char of the previous token (still the
			   tokcurr before we reset it below) to VLPRS
			 */
			*tokcurr->token = VLPRS;
		} else {	/* token doesn't begin with $, #, %, @, or ? */
			vrndx = NIL;
		}
		if ((*token < 0) && ((signed char) *token != SQTPROXY) /* && ((signed char) *token != DQTPROXY) */ ) {
			if (vrndx == NIL) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown var %s", ALMNAM, token+1);
				STATUS = ERROR;
				return;
			}
			if (vrndx == ERROR) {
				logreturn(PIKTLOG, LOG_ERR, ERRID_FN_NAME_RESERVED, __FILE__, __LINE__, fnnam, FALSE, "in %s, attempted use of reserved fn name %s as var", ALMNAM, token+1);
				STATUS = ERROR;
				return;
			}
		}
	}
	ntp = (tok *) mymalloc(sizeof(tok), fnnam);
	ntp->next = (tok *) NULL;
	ntp->token = token;
	switch (exprtyp) {
		case LHS:
			if (! stmcurr->lhsexpr)
				stmcurr->lhsexpr = ntp;
			else
				tokcurr->next = ntp;
			break;
		case RHS:
			if (! stmcurr->rhsexpr)
				stmcurr->rhsexpr = ntp;
			else
				tokcurr->next = ntp;
			break;
		default:
			logreturn(PIKTLOG, LOG_ERR, ERRID_EXPRTYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, bad exprtyp for token %s", ALMNAM, token);
			STATUS = ERROR;
			break;
	}
	tokcurr = ntp;
	return;
}

/* a switch is converted, in effect, into a series of if statements
   internally; for example

      switch #foo
         case 1
            ...
         case 2
            ...
         case 3
            ...
      endsw

   would be converted into

      if #foo == 1
         ...
      endif
      if #foo == 2
         ...
      endif
      if #foo == 3
         ...
      endif

   duptok() duplicates the token sequence "#foo ==" (from the switch
   statement) for each of the newly created if statements
 */
void
duptok (tok *source, tok **target)
{
	static char fnnam[] = "duptok";
	tok *tp, *ntp;

	foreach (tp, source) {
		ntp = (tok *) mymalloc(sizeof(tok), fnnam);
		ntp->next = (tok *) NULL;
		ntp->token = tp->token;
		if (! *target)
			*target = ntp;
		else
			tokcurr->next = ntp;
		tokcurr = ntp;
	}
}

/* register a program statement: initialize data structures; for certain
   statements, create an auxiliary dummy statement (for holding essential
   control structure info); link it to the statement list
 */
void
regstm (int type)
{
	static char fnnam[] = "regstm";
	stm *nsp, *asp;
	register int k;

	nsp = (stm *) mymalloc(sizeof(stm), fnnam);
	nsp->tnext = (stm *) NULL;
	nsp->fnext = (stm *) NULL;
	nsp->type = type;
	nsp->status = ACTIVE;
	nsp->rhsexpr = (tok *) NULL;
	nsp->vrndx = 0;
	nsp->lhsexpr = (tok *) NULL;
	for (k=0; k<=MAXARYDIM; k++)
		nsp->arndx[k] = 0L;
	nsp->arkey = (char *) NULL;
	nsp->logfile = (char *) NULL;
	if ((type == IF) || (type == WHILE) || (type == REPEAT) || (type == FOR) || (type == SWITCH)) {
		asp = (stm *) mymalloc(sizeof(stm), fnnam);
		asp->tnext = (stm *) NULL;
		asp->fnext = (stm *) NULL;
		switch (type) {
			case IF:
				asp->type = ENDIF;
				break;
			case WHILE:
				asp->type = WHILEAUX;
				break;
			case REPEAT:
				asp->type = REPEATAUX;
				break;
			case FOR:
				asp->type = FORAUX;
				break;
			case SWITCH:
				asp->type = ENDSWITCH;
				break;
		}
		asp->rhsexpr = (tok *) NULL;
		asp->vrndx = 0;
		asp->lhsexpr = (tok *) NULL;
		for (k=0; k<=MAXARYDIM; k++)
			asp->arndx[k] = 0L;
		asp->arkey = (char *) NULL;
		asp->logfile = (char *) NULL;
		asp->aux = (stm *) NULL;
		nsp->aux = asp;
	} else
		nsp->aux = (stm *) NULL;
	if (rulcurr && ! rulcurr->statements)
		rulcurr->statements = nsp;
	else if (stmcurr)
		stmcurr->next = stmcurr->tnext = nsp;
	if (stmcurr && (stmcurr->type == CASE)) {	/* previous statement is CASE */
		if ((type == CASE) || (type == DEFAULT))	/* current statement is CASE or DEFAULT */
			stmcurr->fnext = nsp;	/* set prev statement fnext to current statement */
	}
	stmcurr = nsp;
	stmcurr->fnext = nsp->aux;
	return;
}

/* register some history-related info for the current history var
 */
void
reghst (int ndx)
{

	dohistory = TRUE;		/* so, process history file */
	HSTTYP |= VALTYP(ndx);
	NEWHST(ndx) = TRUE;
	return;
}

/* create and initialize the vals data structure
 */
void
initvals (val **values)
{
	static char fnnam[] = "initvals";
	val *valp;
	/* register int i, j; */

	if (*values == (val *) NULL) {
		valp = (val *) mycalloc(1, sizeof(val), fnnam);
		/* RH Linux is unhappy with these initializations; since
		   we call mycalloc() above, they are unnecessary anyway
		   (so long as FORWARD is #defined as 0)
		for (i=0; i<=MAXVARS; i++) {
			for (j=0; j<=9; j++) {
				valp->vals[i][j] = (vl *) NULL;
				valp->nodes[i][j] = 0;
			}
			valp->direction[i] = FORWARD;
		}
		*/
		*values = valp;
	}
	return;
}

/* given a var name, returns its index (1 to 255, or one of the special
   indices)
 */
signed int
varndx (char *name)
{
	static char fnnam[] = "varndx";
	char *nam;
	char *cp;
	register char c;
	register signed int i;

	if (! isalnum((int)name[0]))
		return NIL;
	if (isdigit((int)name[0]))
		return VLPRS;
	i = 0;
	while ((c = *(name+i))) {
		i++;
		if (isalnum((int) c) || (c == '_'))
			continue;
		return VLSTR;
	}
	nam = (char *) mymalloc((strlen(name)+3) * sizeof(char), fnnam);
	sprintf(nam, " %s ", name);
	if ((cp = strstr(nam, "[0]"))) {
		*cp = ' ';
		*(cp+1) = '\0';
	}
	if (strstr(fnnumnames, nam) || strstr(fnstrnames, nam)) {
		free(nam);
		return ERROR;	/* name reserved as function name */
	}
	if (strstr(" inlin inline inputline ", nam)) {
		free(nam);
		return INLINNOW;	/* applies also to @inlin, etc. */
	}
	if (alarmcurr && INPUTBAKFIL && strstr(" outlin outline outputline ", nam)) {
		free(nam);
		return OUTLINNOW;	/* applies also to @outlin, etc. */
	}
	if (strstr(" rdlin rdline readline ", nam)) {
		free(nam);
		return RDLINNOW;
	}
	free(nam);
	for (i=1; i<=VRNDXHI; i++) {
		if (VARNAM(i) && ! strcmp(VARNAM(i), name))
				return i;
	}
	return NIL;
}

/* return a var's dat type (STRING/NUMBER/FILEHANDLE/PROCHANDLE)
 */
int
vardattype (char *name, int namelen, int commentcheck)
{
	static char fnnam[] = "vardattype";
	signed int i;
	long altinpos;
	int lineno = 0;
	char *cp;
	char c;
	int dqt = 0;
	int escape = FALSE;

	if ((i = varndx(name)) == NIL) {
		if (! commentcheck) {
			if (! (altinpos = ftell(altin)))
				logexit(PIKTLOG, LOG_ERR, ERRID_FILE_TELL_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, ftell() failure for altin", ALMNAM);
			lineno = 0;
			while (fgets(altline, MAXBUF, altin) != NULL) {
				if (lineno++ && isalpha((int)altline[0]))
					return ERROR;
				/* look for name in altline */
				cp = altline;
				dqt = 0;
				escape = FALSE;
				while ((c = *cp)) {
					switch (c) {
						case '\\':
							if (! escape) {
								escape = TRUE;
							} else {
								escape = FALSE;
							}
							break;
						case '\"':
							if (! escape) {
								dqt++;
							}
							escape = FALSE;
							break;
						case '#':
						case '$':
							if (    ! escape
							     && even(dqt)
							     && *(cp+1)
							     && ! strncmp(name, cp+1, (unsigned int)namelen)
							     && ! (    isalnum((int)*(cp+1+namelen))
							            || (*(cp+1+namelen) == '_')
							          )
							   ) {
fseek(altin, altinpos, SEEK_SET);
								return (c == '#') ? NUMBER : STRING;
							}
							escape = FALSE;
							break;
						default:
							escape = FALSE;
							break;
					}
					cp++;
				}
			}
			if (! fseek(altin, altinpos, SEEK_SET))
				logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_SEEK_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fseek() failure for altin", ALMNAM);
			logreturn(PIKTLOG, LOG_ERR, ERRID_VAR_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown var %s", ALMNAM, name);
			STATUS = ERROR;
		}
		return ERROR;
	} else if ((i == VLPRS) || (i == INLINNOW) || (i == OUTLINNOW) || (i == RDLINNOW)) {
		return STRING;
	} else if (i == ERROR) {
		return ERROR;
	} else if (DATTYP(i) == NIL) {
		if (! commentcheck) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_DATTYPE_INDETERMINATE, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown dattype for var %s", ALMNAM, name);
			STATUS = ERROR;
		}
		return ERROR;
	} else {
		return DATTYP(i);
	}
}

/* return the next available index in the sequence 1 ... 255
 */
int
freevarslot (void)
{
	register int i;

	for (i=1; i<=MAXVARS; i++) {
		if (! VARNAM(i))
			return i;
	}
	return NIL;
}
